--[[
_ Talented is a complete rewrite of the Talent Frame using a little bit of Ace.

_ Talented makes a lot more assumptions than the original Blizzard code about talents.

_ Talented show all 3 talent tabs at the same time.
_ Talented allows to create and store templates
]]

Talented = AceLibrary("AceAddon-2.0"):new("AceConsole-2.0", "AceDB-2.0", "AceDebug-2.0", "AceHook-2.1", "AceEvent-2.0")

local L = AceLibrary("AceLocale-2.2"):new("Talented")

Talented:RegisterDB("TalentedDB", "TalentedCharDB")
Talented:RegisterDefaults("class", { templates = {} })
Talented:RegisterDefaults("profile", { confirmlearn = true })

function Talented:InitializeOptions()
	self.options = {
		desc = L["Talented - Talent Editor"],
		type = "group",
		args = {
			options = {
				name = L["Options"],
				desc = L["Options for Talented."],
				type = "group",
				args = {
					confirmlearn = {
						name = L["Confirm Learning"],
						desc = L["Ask for user confirmation before learning any talent."],
						type = "toggle",
						get = function() return self.db.profile.confirmlearn end,
						set = function(value) self.db.profile.confirmlearn = value end,
					}
				}
			},
			mode = {
				name = L["Mode"],
				desc = L["Mode of operation."],
				type = "group",
				args = {
					edit = {
						name = L["Edit template"],
						desc = L["Toggle the edition of the template."],
						type = "toggle",
						disabled = function () return self.mode ~= "view" and self.mode ~= "edit" end,
						get = function () return self.mode == "edit" end,
						set = function (value) 
								if self.mode == "edit" then
									self:SetMode("view")
								else
									self:SetMode("edit")
								end
							end,
						order = 100,
					},
					h1 = {
						type = "header",
						order = 150,
					},
					apply = {
						name = L["Apply template"],
						desc = L["Apply the current template to your character."],
						type = "execute",
						order = 200,
						disabled = function () return self.template == self.current end,
						func = function () self:SetMode("apply") end,
						confirm = L["Are you sure that you want this template to be applied ?"],
					},
					delete = {
						name = L["Delete template"],
						desc = L["Delete the current template."],
						type = "execute",
						order = 210,
						disabled = function () return self.template == self.current end,
						func = function () self:SetMode("delete") end,
						confirm = L["Are you sure that you want to delete this template ?"],
					},
					import = {
						name = L["Import template"],
						desc = L["Import a template from Blizzard talent calculator."],
						usage = L["<full url of the template>"],
						type = "text",
						order = 220,
						get = false,
						set = function (url) self:SetTemplate(self:ImportWowTemplate(url)) end,
					},
					export = {
						name = L["Export template"],
						desc = L["Export the template to the current chat target."],
						type = "execute",
						order = 230,
						func = function () self:ExportTemplateToChat() end,
					}
				}
			},
			template = {
				name = L["Template"],
				desc = L["Template"],
				type = "group",
				args = {
					new = {
						name = L["New Template"],
						desc = L["Create a new Template."],
						type = "group",
						order = 100,
						args = {
							empty = {
								name = L["New empty template"],
								desc = L["Create a new template from scratch."],
								type = "execute",
								order = 100,
								func = function() self:SetTemplate(self:CreateEmptyTemplate()) end,
							},
							h1 = {
								type = "header",
								order = 150,
							},
							-- copy from existing templates here
							current = {
								name = L["Copy from current"],
								desc = L["Create a new template from the current."],
								type = "execute",
								order = 300,
								func = function() self:SetTemplate(self:CopyTemplate(self.current)) end,
							},
						},
					},
					h1 = {
						type = "header",
						order = 150,
					},
					-- copy from existing templates here
					current = {
						name = L["Current template"],
						desc = L["Select the current template."],
						type = "execute",
						order = 300,
						func = function() self:SetTemplate(self.current) end,
					},
				}
			}
		},
	}
	
	do
		local src = self.db.char.templates
		local dst = self.db.class.templates
		if src then
			for name, template in pairs(src) do
				if not dst[name] then
					dst[name] = template
					src[name] = nil
				end
			end
		end
	end
	
	for _, template in pairs(self.db.class.templates) do
		self:AddTemplateToOptions(template)
	end
	self:RegisterChatCommand(L["CHAT_COMMANDS"], self.options.args.options)
end

function Talented:GetMode()
	return self.mode
end

function Talented:ApplyCurrentTemplate()
	local template = self.template
	local count = 0
	-- check if enough talent points are available
	local available = UnitCharacterPoints("player")
	for tab, talents in ipairs(self.talents) do
		for index, info in ipairs(talents.infos) do
			local delta = template[tab][index] - self.current[tab][index]
			if delta > 0 then
				count = count + delta
			end
		end
	end
	if count > available then
		self:Print(L["Sorry, I can't apply this template because you don't have enough talent points available (need %d)!"], count)
		self.mode = "view"
		self:UpdateView()
	else
		-- let's go !
		self:Print(L["Please wait while I set your talents..."])
		self:EnableUI(false)
		self:CHARACTER_POINTS_CHANGED()
	end
end

function Talented:SetMode(mode)
	if self.template == self.current and mode ~= "edit" then
		mode = "view"
	end
	if self.mode ~= mode then
		self.mode = mode
		if mode == "apply" then
			self:ApplyCurrentTemplate()
		elseif mode == "delete" then
			self:DeleteCurrentTemplate()
		end
		self:UpdateView()
	end
end

function Talented:OnInitialize()
	self.talents = {}
	self.templaterefs = {}
	
	self.mode = "view"
	
	self:InitializeOptions()
end

function Talented:RemoveTemplateFromOptions(template)
	local key = self.templaterefs[template]
	if key then
		self.options.args.template.args[key] = nil
		self.options.args.template.args.new.args[key] = nil
		self.templaterefs[template] = nil
	end
end

function Talented:DeleteCurrentTemplate()
	local template = self.template
	if template == self.current then return end
	self:RemoveTemplateFromOptions(template)
	self.db.class.templates[template.name] = nil
	self:SetTemplate(self.current)
end

function Talented:UpdateTemplateName(template, newname)
	if self.db.class.templates[newname] or
			self.mode == "apply" or
			self.current == template then
		return
	end
	
	local oldname = template.name
	template.name = newname
	self.db.class.templates[newname] = template
	self.db.class.templates[oldname] = nil
	local key = self.templaterefs[template]
	if key then
		local format = string.format
		self.options.args.template.args[key].name = template.name
		self.options.args.template.args[key].desc = format(L["Select %s"], template.name)
		self.options.args.template.args.new.args[key].name = format(L["Copy from %s"], template.name)
		self.options.args.template.args.new.args[key].desc = format(L["Create a new template from %s."], template.name)
	end
end

function Talented:AddTemplateToOptions(template)
	local format = string.format
	
	self.db.class.templates[template.name] = template
	local count, opt, name = 1, self.options.args.template.args
	while true do
		name = format("template%d", count)
		if not opt[name] then break end
		count = count + 1
	end
	self.templaterefs[template] = name
	opt[name] = {
		name = template.name,
		desc = format(L["Select %s"], template.name),
		type = "execute",
		order = 200,
		func = function () self:SetTemplate(template) end,
	}
	self.options.args.template.args.new.args[name] = {
		name = format(L["Copy from %s"], template.name),
		desc = format(L["Create a new template from %s."], template.name),
		type = "execute",
		order = 200,
		func = function() self:SetTemplate(self:CopyTemplate(template)) end,
	}
end

do
	local function new(templates, name)
		local count = 0
		local template = {
			name = name
		}
		while templates[template.name] do
			count = count + 1
			template.name = string.format(L["%s (%d)"], name, count)
		end
		return template
	end
	
	function Talented:CopyTemplate(src)
		local dst = new(self.db.class.templates, string.format(L["Copy of %s"], src.name))
		
		for tab, talents in ipairs(self.talents) do
			local s, d = src[tab], {}
			dst[tab] = d
			for index in ipairs(talents.infos) do
				d[index] = s[index]
			end
		end

		self:AddTemplateToOptions(dst)
		return dst
	end

	function Talented:CreateEmptyTemplate()
		local template = new(self.db.class.templates, L["Empty"])
		
		for tab, talents in ipairs(self.talents) do
			local t = {}
			template[tab] = t
			for index in ipairs(talents.infos) do
				t[index] = 0
			end
		end

		self:AddTemplateToOptions(template)
		return template
	end
	
	function Talented:ImportWowTemplate(url, force)
		local s, _, class, src = url:find(".*/(.*)/talents.html%?(%d+)$")
		if not s then
			self:Print(L["\"%s\" does not appear to be a valid URL!"], url)
			return
		end
		local _, pclass = UnitClass("player")
		local count, len = 0, src:len()
		for tab, talents in ipairs(self.talents) do
			count = count + #talents.infos
		end
		if not force and (pclass:lower() ~= class or count ~=len) then
			self:Debug("count = %d, len = %d, class = %s, pclass = %s", count, len, class, pclass)
			self:Print(L["\"%s\" does not appear to be a valid URL!"], url)
			return
		end
		
		local dst = new(self.db.class.templates, L["Imported"])
		count = 1
		for tab, talents in ipairs(self.talents) do
			local t = {}
			dst[tab] = t
			for index in ipairs(talents.infos) do
				t[index] = tonumber(src:sub(count, count))
				count = count + 1
				if count > len then break end
			end
			if not self:ValidateTalentBranch(dst, tab) then
				self:Print(L["The given template is not a valid one! branch %d is invalid."], tab)
				return
			end
		end

		self:AddTemplateToOptions(dst)
		return template
	end
end

function Talented:SetTemplate(template)
	if not template then template = self.current end
	self.template = template
	self.mode = "view"
	self:UpdateView()
end

function Talented:OnEnable()
	self:Hook("ToggleTalentFrame", true)
	self:SecureHook("UpdateTalentButton")
	self:RegisterEvent("CHARACTER_POINTS_CHANGED")
end

function Talented:CHARACTER_POINTS_CHANGED()
	self:UpdateCurrentTemplate()
	self:UpdateView()
	if self.mode == "apply" then
		local cp = UnitCharacterPoints("player")
		
		local reset = function(...)
			if select("#", ...) > 0 then
				self:Print(...)
			end
			self.template = self.current
			self.mode = "view"
			self:UpdateView()
			self:EnableUI(true)
		end

		local current = self.current
		local template = self.template
		for tab, talents in ipairs(self.talents) do
			local ctab = current[tab]
			local ttab = template[tab]
			for index in ipairs(talents.infos) do
				local current = ctab[index]
				local target = ttab[index]
				if current < target then
					if cp == 0 then
						reset(L["Error while applying talents! Not enough character points!"])
						return
					end
					self:RegisterEvent("CHARACTER_POINTS_CHANGED")
					self:LearnTalent(tab, index)
					return
				end
			end
		end
		reset(L["Template applied successfully, %d talent points remaining."], UnitCharacterPoints("player"))
	end
end

function Talented:UpdateTalentButton()
	if self.db.profile.donthide and UnitLevel("player") < 10 then
		TalentMicroButton:Show()
		QuestLogMicroButton:SetPoint("BOTTOMLEFT", "TalentMicroButton", "BOTTOMRIGHT", -2, 0)
	end
end

function Talented:ToggleTalentFrame()
	local frame = self.base
	if not frame or not frame:IsVisible() then
		self:Update()
		ShowUIPanel(self.base)
	else
		HideUIPanel(frame)
	end
end

function Talented:Update()
	self:CreateBaseFrame()
	self:UpdateCurrentTemplate()
	if not self.template then
		self:SetTemplate(self.current)
	end
	self:UpdateView()
end


function Talented:ExportTemplateToChat()
	local write
	do
		local edit = DEFAULT_CHAT_FRAME.editBox
		local SendChatMessage = SendChatMessage
		local type = edit:GetAttribute("chatType")
		local lang = edit.language
		if type == "WHISPER" then
			local target = edit:GetAttribute("tellTarget")
			write = function (text, ...)
				if text:find("%", 1, true) then text = text:format(...) end
				SendChatMessage(text, type, lang, target)
			end
		elseif type == "CHANNEL" then
			local channel = edit:GetAttribute("channelTarget")
			write = function (text, ...)
				if text:find("%", 1, true) then text = text:format(...) end
				SendChatMessage(text, type, lang, channel)
			end
		else
			write = function (text, ...)
				if text:find("%", 1, true) then text = text:format(...) end
				SendChatMessage(text, type, lang)
			end
		end
	end
	write(L["Talented - Template \"%s\" - %s :"], self.template.name, UnitClass("player"))
	for tab, talents in ipairs(self.talents) do
		local branchwritten = false
		for index, talent in ipairs(talents.infos) do
			local value = self.template[tab][index]
			if value > 0 then
				local ranks = talent.ranks
				if not branchwritten then
					write(L["%s :"], talents.name)
					branchwritten = true
				end
				if ranks == 1 then
					write(L["_ %s"], talent.name)
				else
					write(L["_ %s (%d/%d)"], talent.name, value, ranks)
				end
			end
		end
	end
end
